{$COPERATORS ON}

unit glKernedFont;

interface

uses
	gl,
	glu;
	
type
	tglKernedChar=array[0..5] of shortInt;
	tglKernedTable=array[32..127] of tglKernedChar;
	tglKernedHeader=record
		formatRevision,
		bufferWidth,bufferHeight,
		charWidth,charHeight:dword;
		kerningTable:tglKernedTable;	
		reserved:array[0..127] of byte;
	end;
	
	tglKernedFont=object
		name:string;
		texture:gluInt;
		info:tglKernedHeader;
		renderX,renderY,renderZ,
		charGlHeight,
		charAddWidth,charAddHeight,charSpacing,
		charScale,charScaleWidth,charScaleHeight:glFloat;
		lastKern:tglKernedChar;
		constructor init(fontName:string; glHeight:glFloat);
		procedure scale(s:glFloat);
		procedure setposition(x,y,z:glFloat);
		function stringSize(st:string):glFloat;
		procedure writeChar(ch:char);
		procedure writeString(st:string);
		procedure writeStringCentered(st:string);
		procedure writeStringRight(st:string);
		destructor term;
	end;
	
implementation

function kernCompare(prev,this:tglKernedChar):glFloat;
var
	t:word;
	n,test:glFloat;
begin
	n:=-999;
	for t:=0 to 2 do begin
		test:=prev[t+3]+this[t];
		if test>n then n:=test;
	end;
	kernCompare:=n;
end;

constructor tglKernedFont.init(fontName:string; glHeight:glFloat);
var
	f:file;
	fBuffer,fPoint,fBufferEnd,
	pBuffer:^byte;
	pPoint:^word;
	t,b:word;
	fBufferSize,pBufferSize:dword;
begin
	name:=fontName;
	
	assign(f,fontName+'.glkfont');
	reset(f,1);
		blockread(f,info,sizeof(info));
		fBufferSize:=filesize(f)-sizeof(info);	
		getmem(fBuffer,fBufferSize);
		blockread(f,fBuffer^,fBufferSize);
	close(f);
	
	{ texture2f scale values }
	with info do begin
		charAddWidth:=(charWidth-1)/bufferWidth;
		charAddHeight:=(charHeight-1)/bufferHeight;
		pBufferSize:=trunc(bufferWidth)*trunc(bufferHeight)*2;
		getmem(pBuffer,pbufferSize);
	end;
	
	{ vector scale values	}
	charGlHeight:=glHeight;
	scale(1);	

	fPoint:=fBuffer;
	fBufferEnd:=fPoint+fBufferSize;
	pPoint:=pointer(pBuffer);
	
	while not(fpoint>=fBufferEnd) do begin
		case fPoint^ of 
			0,255:begin
				b:=$FF or (fPoint^ shl 8);
				inc(fPoint);
				t:=fPoint^;
				while (t>0) do begin
					pPoint^:=b;
					inc(pPoint);
					dec(t);
				end;
			end;
			else begin
				pPoint^:=$FF or (fPoint^ shl 8);
				inc(pPoint);
			end;
		end;
		inc(fPoint);
	end;
	
	freeMem(fBuffer);
	
	glGenTextures(1,@texture);
	glBindTexture(GL_TEXTURE_2D,texture);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri (GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	gluBuild2DMipmaps(
		GL_TEXTURE_2D,2,
		info.bufferWidth,info.bufferHeight,
		GL_LUMINANCE_ALPHA,
		GL_UNSIGNED_BYTE,
		pBuffer
	);
	glBindTexture(GL_TEXTURE_2D,0); // disable it
	freeMem(pBuffer);

	renderX:=0;
	renderY:=0;
	charspacing:=3;
end;

procedure tglKernedFont.scale(s:glFloat);
var
	t:word;
begin
	charScale:=(charGlHeight/info.charHeight)*s;
	charScaleWidth:=info.charWidth*charScale;
	charScaleHeight:=info.charHeight*charScale;
	for t:=0 to 5 do lastKern[t]:=0;
end;

procedure tglKernedFont.setposition(x,y,z:glFloat);
var
	t:word;
begin
	renderX:=x;
	renderY:=y;
	renderZ:=z;
	for t:=0 to 5 do lastKern[t]:=0;
end;

function tglKernedFont.stringSize(st:string):glFloat;
var
	kern:tglKernedChar;
	calcX,r:glFloat;
	t:word;
begin
	for t:=0 to 5 do kern[t]:=0;
	calcX:=0;
	if (length(st)>0) then begin
		for t:=1 to length(st) do begin
			calcX+=(
				(
					kernCompare(kern,info.kerningTable[ord(st[t])]) +
					charSpacing
				) *	charScale
			);
			kern:=info.kerningTable[ord(st[t])];
		end;
		r:=0;
		for t:=3 to 5 do begin
			if kern[t]>r then r:=kern[t];
		end;
		calcX+=r*charScale;
	end;
	stringSize:=calcX;
end;

procedure tglKernedFont.writeChar(ch:char);
var
	vsX,vsY,veX,veY,
	tsX,tsY,teX,teY:glFloat;
begin
	case ch of
		#32..#127:begin
			{	scale the character positions to 0..1	}
			tsX:=(ord(ch) mod 32)/32; 
			tsY:=((ord(ch) div 32)-1)/3;
			teX:=tsX+charAddWidth;
			teY:=tsY+charAddHeight;
			
			{ perform kerning }
			renderX+=(
				kernCompare(lastKern,info.kerningTable[ord(ch)]) +
				charSpacing
			)*charScale;
			
			vsX:=renderX;
			vsY:=renderY;
			veX:=renderX+charScaleWidth;
			veY:=renderY+charScaleHeight;
			
			glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
			glBindTexture(GL_TEXTURE_2D,texture);
			glBegin(GL_QUADS);
				glTexCoord2f(tsX,tsY);
				glVertex3f(vsX,vsY,renderZ);
				glTexCoord2f(teX,tsY);
				glVertex3f(veX,vsY,renderZ);
				glTexCoord2f(teX,teY);
				glVertex3f(veX,veY,renderZ);
				glTexCoord2f(tsX,teY);
				glVertex3f(vsX,veY,renderZ);
			glEnd;
			glBindTexture(GL_TEXTURE_2D,0);
			
			lastKern:=info.kerningTable[ord(ch)];
		end;
	end;
end;

procedure tglKernedFont.writeString(st:string);
var
	t:word;
begin
	{ use while so that zero length is unprocessed! }
	t:=0;
	while (t<length(st)) do begin
		inc(t);
		writeChar(st[t]);
	end;
end;

procedure tglKernedFont.writeStringCentered(st:string);
begin
	renderX-=stringSize(st)/2;
	renderY-=(info.charHeight*charScale)/3; // one third == center of capsline and baseline
	writeString(st);
end;
	
procedure tglKernedFont.writeStringRight(st:string);
begin
	renderX-=stringSize(st);
	writeString(st);
end;

destructor tglKernedFont.term;
begin
	glDeleteTextures(1,@texture);
end;

end.